import Logger from "@home-gallery/logger";
import { through } from "@home-gallery/stream";
import { serialize, createHash } from "@home-gallery/common";
const log = Logger("plugin.databaseMapperFactory");
const validateMapper = (mapper) => {
  return typeof mapper.name == "string" && typeof mapper.mapEntry == "function";
};
const getDatabaseMapper = (plugin, factory, disabledMappers) => {
  const mappers = factory.getDatabaseMappers();
  if (!mappers?.length) {
    return [];
  }
  const mapperEntries = [];
  for (let databaseMapper of mappers) {
    if (disabledMappers.includes(databaseMapper.name)) {
      log.info(`Disable database mapper ${databaseMapper.name} from ${plugin.name}`);
      continue;
    } else if (!validateMapper(databaseMapper)) {
      log.warn(`Invalid database mapper ${databaseMapper.name} from plugin ${plugin.name}. Ignore it`);
      continue;
    }
    mapperEntries.push({
      databaseMapper,
      plugin
    });
  }
  return mapperEntries.sort((a, b) => {
    const aOrder = a.databaseMapper.order || 1;
    const bOrder = b.databaseMapper.order || 1;
    return aOrder <= bOrder ? -1 : 1;
  });
};
const createStream = (mapperEntries, config, updated) => {
  const stream = through((entry, _, cb) => {
    let media = {
      id: entry.sha1sum,
      hash: "",
      type: entry.type,
      date: "",
      updated,
      files: [],
      previews: [],
      plugin: {}
    };
    for (let mapperEntry of mapperEntries) {
      const mapper = mapperEntry.databaseMapper;
      try {
        const result = mapper.mapEntry?.(entry, media, config);
        if (result && typeof result == "object") {
          media = result;
        }
      } catch (err) {
        log.warn(err, `Database mapper ${mapper.name} from plugin ${mapperEntry.plugin.name} failed: ${err}. Ignore`);
      }
    }
    media.hash = createHash(serialize(media, "hash"));
    cb(null, media);
  });
  return stream;
};
const createDatabaseMapperStream = (plugins, config, updated) => {
  const entries = [];
  const disabledMappers = config.pluginManager?.disabledMappers || [];
  for (let plugin of plugins) {
    if (!plugin.factory?.getDatabaseMappers) {
      continue;
    }
    const databaseEntries = getDatabaseMapper(plugin.plugin, plugin.factory, disabledMappers);
    entries.push(...databaseEntries);
  }
  if (!entries.length) {
    throw new Error(`No database mapper found. At least one is required`);
  }
  const stream = createStream(entries, config, updated);
  return {
    stream,
    entries
  };
};
export {
  createDatabaseMapperStream,
  createStream,
  getDatabaseMapper
};
//# sourceMappingURL=databaseMapperStreamFactory.js.map
