import { Transform } from "stream";
import { through } from "@home-gallery/stream";
import Logger from "@home-gallery/logger";
const log = Logger("plugin.extractorStreamFactory");
class ExtractorStreamFactory {
  manager;
  storage;
  plugins;
  constructor(manager, storage, plugins) {
    this.manager = manager;
    this.storage = storage;
    this.plugins = plugins;
  }
  #functionToStream(task, extractor) {
    return through(function transform(entry, _, cb) {
      Promise.resolve(task(entry)).then(() => cb(null, entry)).catch((err) => {
        cb(new Error(`Extractor function task ${extractor.name} for phase ${extractor.phase} failed `, { cause: err }));
      });
    });
  }
  #testAndTaskToStream(task, plugin) {
    const test = task.test ? task.test : () => true;
    const end = task.end ? task.end : () => Promise.resolve();
    return through(function transform(entry, _, cb) {
      if (!test(entry)) {
        return cb(null, entry);
      }
      task.task(entry).then(() => cb(null, entry)).catch((err) => {
        cb(new Error(`Extractor task ${plugin.name} for phase ${plugin.phase} failed`, { cause: err }));
      });
    }, function flush(cb) {
      end().then(cb, cb);
    });
  }
  async getExtractorStreamsFrom(plugin, extractors) {
    if (!extractors?.length) {
      return [];
    }
    const config = this.manager.getConfig();
    const streams = [];
    for (let extractor of extractors) {
      await extractor.create(this.storage).then((task) => {
        if (task instanceof Transform) {
          const logError = (err) => log.warn(err, `Extractor transform task ${extractor.name} for phase ${extractor.phase} failed`);
          streams.push({ stream: task.on("error", logError), extractor, plugin });
        } else if (typeof task == "function") {
          streams.push({ stream: this.#functionToStream(task, extractor), extractor, plugin });
        } else if (typeof task.task == "function") {
          streams.push({ stream: this.#testAndTaskToStream(task, extractor), extractor, plugin });
        } else {
          throw new Error(`Invalid task. Expected Transform, function or object with task function`);
        }
      }).catch((err) => {
        log.warn(err, `Failed to create extractor task of ${extractor.name} for phase ${extractor.phase}: ${err}. Skip it`);
      });
    }
    return streams;
  }
  async getExtractorStreams() {
    const disabledExtractors = this.manager.getConfig().pluginManager?.disabledExtractors || [];
    const t0 = Date.now();
    const streams = [];
    for (let plugin of this.plugins) {
      if (!plugin.factory?.getExtractors) {
        continue;
      }
      const extractors = plugin.factory.getExtractors().filter((extractor) => {
        if (!disabledExtractors.includes(extractor.name)) {
          return true;
        }
        log.info(`Disable extractor ${extractor.name} from ${plugin.plugin.name}`);
        return false;
      });
      const pluginStreams = await this.getExtractorStreamsFrom(plugin.plugin, extractors);
      streams.push(...pluginStreams);
    }
    log.debug(t0, `Loaded ${streams.length} extractor tasks from ${this.plugins.length} plugins`);
    const tearDown = async () => {
      const tearDownExtractors = streams.filter((stream) => stream.extractor?.tearDown).map((stream) => stream.extractor).reverse();
      if (!tearDownExtractors.length) {
        return;
      }
      const t02 = Date.now();
      for (let plugin of tearDownExtractors) {
        await plugin.tearDown().catch((err) => log.warn(err, `Failed to tear down plugin ${plugin.name}: ${err}`));
      }
      log.debug(t02, `All ${tearDownExtractors.length} extractors where teared down: ${tearDownExtractors.map((p) => p.name).join(", ")}`);
    };
    return [streams, tearDown];
  }
}
export {
  ExtractorStreamFactory
};
//# sourceMappingURL=extractorStreamFactory.js.map
