import fs from "fs/promises";
import path from "path";
import Logger from "@home-gallery/logger";
import { resolveValid } from "./load/resolve.js";
import { resolvePackageEntry } from "./load/pkg.js";
import { ExtractorStreamFactory } from "./extractorStreamFactory.js";
import { createDatabaseMapperStream } from "./databaseMapperStreamFactory.js";
import { QueryExecutor } from "../query/queryExecutor.js";
class PluginManager {
  plugins = [];
  config;
  context;
  managerConfig;
  log;
  queryExecutor;
  constructor(config = {}, context = {}) {
    this.config = config;
    this.context = context;
    this.managerConfig = config.pluginManager || {};
    this.log = Logger("pluginManager");
    this.queryExecutor = new QueryExecutor();
  }
  getApiVersion() {
    return "0.7";
  }
  getConfig() {
    return this.config;
  }
  createLogger(module) {
    return Logger(module);
  }
  getContext() {
    return this.context;
  }
  async importPlugin(file) {
    const module = await import(file);
    const plugin = module.default ?? module;
    this.validatePluginProps(file, plugin);
    return this.addPlugin(file, plugin);
  }
  validatePluginProps(file, plugin) {
    if (typeof plugin != "object") {
      throw new Error(`Invalid plugin ${file}: No object was exported`);
    }
    const requiredProps = [
      { name: "name", type: "string" },
      { name: "version", type: "string" },
      { name: "initialize", type: "function" }
    ];
    const missingProps = requiredProps.map((p) => p.name).filter((name) => typeof plugin[name] == "undefined");
    if (missingProps.length) {
      throw new Error(`Invalid plugin ${file}: Missing properties ${missingProps.join(",")}`);
    }
    const invalidProps = requiredProps.filter((prop) => typeof plugin[prop.name] != prop.type);
    if (invalidProps.length) {
      throw new Error(`Invalid plugin ${file}: Incompatible property types: ${invalidProps.map((p) => `${p.name} should be a ${p.type}`).join(" and ")}`);
    }
  }
  async addPlugin(file, plugin) {
    const { name } = plugin;
    const nameExists = this.plugins.find((ctx) => ctx.plugin.name == name);
    if (nameExists) {
      throw new Error(`Duplicate plugin name ${plugin.name}`);
    }
    if (this.managerConfig.disabled?.includes(name)) {
      this.log.debug(`Skip loading disabled plugin ${plugin.name}`);
      return;
    }
    this.plugins.push({
      file,
      plugin,
      loaded: false
    });
    return plugin;
  }
  async getPackageEntry(dir) {
    const data = await fs.readFile(path.join(dir, "package.json"), "utf-8");
    let pkg;
    try {
      pkg = JSON.parse(data);
    } catch (err) {
      throw new Error(`Invalid package.json format: ${err}`);
    }
    const main = resolvePackageEntry(pkg);
    if (main) {
      return path.resolve(dir, main);
    }
    const files = await fs.readdir(dir);
    const index = ["index.js", "index.mjs", "index.cjs"].find((index2) => files.includes(index2));
    if (index) {
      return path.resolve(dir, index);
    }
    throw new Error(`Could not resolve package entry for ${path.join(dir, "package.json")}`);
  }
  async loadPlugin(dirOrFile) {
    const stat = await fs.stat(dirOrFile);
    if (stat.isFile() && dirOrFile.match(/\.[mc]?js$/)) {
      return this.importPlugin(path.resolve(dirOrFile));
    } else if (stat.isDirectory()) {
      const files = await fs.readdir(dirOrFile);
      if (files.includes("package.json")) {
        const packageEntry = await this.getPackageEntry(dirOrFile);
        return this.importPlugin(packageEntry);
      }
      const indexFile = ["index.js", "index.mjs", "index.cjs"].find((index) => files.includes(index));
      if (!indexFile) {
        throw new Error(`Could not find index file in plugin directory ${dirOrFile}`);
      }
      return this.importPlugin(path.resolve(dirOrFile, indexFile));
    } else {
      throw new Error(`Invalid plugin location ${dirOrFile}`);
    }
  }
  async loadPluginDir(dir) {
    const t0 = new Date();
    const dirPlugins = [];
    const pluginDir = path.resolve(dir);
    const exists = await fs.access(dir).then(() => true, () => false);
    if (!exists) {
      this.log.warn(`Plugin dir ${pluginDir} does not exists. Skip it`);
      return;
    }
    const files = await fs.readdir(dir);
    for (let file of files) {
      const pluginFile = path.resolve(dir, file);
      const stat = await fs.stat(pluginFile);
      if (!stat.isDirectory() && (!stat.isFile() || !pluginFile.match(/\.[mc]?js$/))) {
        continue;
      }
      await this.loadPlugin(pluginFile).then((plugin) => plugin && dirPlugins.push(plugin)).catch((err) => this.log.warn(err, `Failed to load plugin from ${pluginFile}`));
    }
    if (!dirPlugins.length) {
      this.log.info(t0, `No plugins found in directory ${dir}`);
    } else {
      this.log.debug(t0, `Loaded ${dirPlugins.length} plugins from directory ${dir}: ${dirPlugins.map((p) => p.name).join(", ")}`);
    }
  }
  validateFactory(factory) {
    if (typeof factory != "object") {
      throw new Error(`Invalid factory`);
    }
    const fns = [
      "getExtractors",
      "getDatabaseMappers",
      "getQueryPlugins"
    ];
    const foundFn = fns.filter((fn) => typeof factory[fn] != "undefined");
    if (!foundFn.length) {
      throw new Error(`No module factory method found. At least on is required of: ${fns.join(", ")}`);
    }
    for (const fn of foundFn) {
      if (typeof factory[fn] != "function") {
        throw new Error(`Invalid module factory method ${fn}. ${fn} is not a function`);
      }
    }
    return factory;
  }
  async initializePlugins() {
    const orderedPlugins = resolveValid(this.plugins);
    for (let ctx of orderedPlugins) {
      await ctx.plugin.initialize(this).then(this.validateFactory).then((factory) => {
        const queryPlugins = factory.getQueryPlugins ? factory.getQueryPlugins() : [];
        this.queryExecutor.addQueryPlugins(queryPlugins);
        ctx.factory = factory;
        ctx.loaded = true;
      }).catch((err) => {
        this.log.error(err, `Failed to initialize plugin ${ctx.plugin.name} (from ${ctx.file}): ${err}`);
      });
    }
  }
  getPlugin(name) {
    const ctx = this.plugins.find((ctx2) => ctx2.loaded && ctx2.plugin.name == name);
    if (ctx) {
      return ctx.plugin;
    }
  }
  getPlugins() {
    return this.plugins.filter((ctx) => ctx.loaded && ctx.factory).map((ctx) => ctx.plugin);
  }
  getModuleFactoryFor(name) {
    return this.plugins.find((p) => p.plugin.name == name && p.loaded)?.factory;
  }
  async getExtractorStreams(storage) {
    const loadedPlugins = this.plugins.filter((ctx) => ctx.loaded && ctx.factory);
    const streamFactory = new ExtractorStreamFactory(this, storage, loadedPlugins);
    return streamFactory.getExtractorStreams();
  }
  async getDatabaseMapperStream(updated) {
    const loadedPlugins = this.plugins.filter((ctx) => ctx.loaded && ctx.factory);
    return createDatabaseMapperStream(loadedPlugins, this.getConfig(), updated);
  }
  async executeQuery(entries, query, context) {
    return this.queryExecutor.execute(entries, query, context);
  }
}
export {
  PluginManager
};
//# sourceMappingURL=manager.js.map
